/*
 * Copyright (c) 2016-2021 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>

#include <satlab/srs4.h>

#define PROP_CLIENT_SUBSYS_SPEC_CAPSULE	"propclient.subsys_spec"

/* SRS4 object */
typedef struct {
	PyObject_HEAD
	uint8_t node;
	uint32_t timeout;
	PyObject *prop;
	PyObject *csp;
} SRS4Object;

static PyObject *SRS4_gwdt_reset(SRS4Object *self, PyObject *args, PyObject *kwds)
{
	int ret;

	static char *kwlist[] = {"timeout", NULL};
	uint32_t timeout = self->timeout;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "|I", kwlist, &timeout))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	ret = sl_srs4_gwdt_reset(self->node, timeout, SL_SRS4_GWDT_RESET_KEY);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_TimeoutError, "Failed to reset GWDT: %d\n", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static PyObject *SRS4_gwdt_counter(SRS4Object *self, PyObject *args, PyObject *kwds)
{
	int ret;
	uint32_t counter;

	static char *kwlist[] = {"timeout", NULL};
	uint32_t timeout = self->timeout;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "|I", kwlist, &timeout))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	ret = sl_srs4_gwdt_counter(self->node, timeout, &counter);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_TimeoutError, "Failed to read GWDT: %d\n", ret);
		return NULL;
	}

	return PyLong_FromLong((long)counter);
}

static PyObject *SRS4_shell_run(SRS4Object *self, PyObject *args, PyObject *kwds)
{
	int ret;

	static char *kwlist[] = {"cmd", "timeout", NULL};
	const char *cmd;
	uint32_t timeout = self->timeout;

	if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|I", kwlist, &cmd, &timeout))
		return NULL;

	Py_BEGIN_ALLOW_THREADS
	ret = sl_srs4_shell_run(self->node, timeout, cmd);
	Py_END_ALLOW_THREADS
	if (ret < 0) {
		PyErr_Format(PyExc_TimeoutError, "Failed to run command: %d\n", ret);
		return NULL;
	}

	Py_RETURN_NONE;
}

static int SRS4_init(SRS4Object *self, PyObject *args, PyObject *kwds)
{
	int ret = -1;
	self->node = SL_SRS4_DEFAULT_ADDRESS;
	self->timeout = 1000;

	static char *kwlist[] = {"node", "timeout", NULL};
	if (!PyArg_ParseTupleAndKeywords(args, kwds, "|bI", kwlist, &self->node, &self->timeout))
		goto err;

	/* Get reference to propclient module */
	PyObject *propclient = PyImport_ImportModule("satlab.propclient");
	if (!propclient)
		goto err;

	/* Get reference to PropNode class */
	PyObject *PropNode = PyObject_GetAttrString(propclient, "PropNode");
	if (!PropNode)
		goto err_prop_node;

	/* Create capsule object with props */
	PyObject *props_capsule = PyCapsule_New((void *)&srs4_props, PROP_CLIENT_SUBSYS_SPEC_CAPSULE, NULL);
	if (!props_capsule)
		goto err_prop_capsule;

	/* Create PropNode object by calling constructor */
	self->prop = PyObject_CallFunction(PropNode, "bbOI", self->node, SL_PROP_DEFAULT_PORT, props_capsule, self->timeout);
	if (!self->prop)
		goto err_prop_ctor;

	/* Get reference to propclient module */
	PyObject *pycsp = PyImport_ImportModule("csp");
	if (!pycsp)
		goto err_csp;

	/* Get reference to PropNode class */
	PyObject *CSPNode = PyObject_GetAttrString(pycsp, "CSPNode");
	if (!CSPNode)
		goto err_csp_node;

	/* Create PropNode object by calling constructor */
	self->csp = PyObject_CallFunction(CSPNode, "b", self->node);
	if (!self->csp)
		goto err_csp_ctor;

	ret = 0;

err_csp_ctor:
	Py_DECREF(CSPNode);
err_csp_node:
	Py_DECREF(pycsp);
err_csp:
err_prop_ctor:
	Py_DECREF(props_capsule);
err_prop_capsule:
	Py_DECREF(PropNode);
err_prop_node:
	Py_DECREF(propclient);
err:
	return ret;
}

static PyMemberDef SRS4_members[] = {
	{"prop", T_OBJECT_EX, offsetof(SRS4Object, prop), 0, "PropNode"},
	{"csp",  T_OBJECT_EX, offsetof(SRS4Object, csp),  0, "CSPNode"},
	{NULL}  /* Sentinel */
};

static PyMethodDef SRS4_methods[] = {
	{"gwdt_reset",   (PyCFunction)SRS4_gwdt_reset,   METH_VARARGS | METH_KEYWORDS, "Reset ground watchdog timer."},
	{"gwdt_counter", (PyCFunction)SRS4_gwdt_counter, METH_VARARGS | METH_KEYWORDS, "Get ground watchdog timer value."},
	{"shell_run",    (PyCFunction)SRS4_shell_run,    METH_VARARGS | METH_KEYWORDS, "Run shell command."},
	{NULL}  /* Sentinel */
};

static PyTypeObject SRS4Type = {
	PyVarObject_HEAD_INIT(NULL, 0)
	.tp_name = "srs4client.SRS4",
	.tp_doc = "SRS4 object",
	.tp_basicsize = sizeof(SRS4Object),
	.tp_itemsize = 0,
	.tp_flags = Py_TPFLAGS_DEFAULT,
	.tp_new = PyType_GenericNew,
	.tp_init = (initproc) SRS4_init,
	.tp_members = SRS4_members,
	.tp_methods = SRS4_methods,
};

static struct PyModuleDef srs4client_module = {
	.m_base = PyModuleDef_HEAD_INIT,
	.m_name = "srs4client",
	.m_size = -1,
};

PyMODINIT_FUNC PyInit_srs4client_python(void)
{
	if (PyType_Ready(&SRS4Type) < 0)
		return NULL;

	PyObject *m = PyModule_Create(&srs4client_module);
	if (!m)
		return NULL;

	/* Add SRS4 object */
	Py_INCREF(&SRS4Type);
	if (PyModule_AddObject(m, "SRS4", (PyObject *) &SRS4Type) < 0) {
		Py_DECREF(&SRS4Type);
		Py_DECREF(m);
		return NULL;
	}

	PyModule_AddIntConstant(m, "DEFAULT_ADDRESS", SL_SRS4_DEFAULT_ADDRESS);

	return m;
}
