/*
 * Copyright (c) 2016-2022 Satlab A/S <satlab@satlab.com>
 * All rights reserved. Do not distribute without permission.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>

#include <slash/slash.h>

#include <csp/csp.h>
#include <csp/drivers/usart.h>
#include <csp/drivers/can_socketcan.h>
#include <csp/crypto/csp_xtea.h>
#include <csp/crypto/csp_hmac.h>
#include <csp/interfaces/csp_if_lo.h>
#include <csp/interfaces/csp_if_can.h>
#include <csp/interfaces/csp_if_kiss.h>
#include <csp/interfaces/csp_if_mcast.h>
#include <csp/interfaces/csp_if_slgnd.h>
#include <csp/interfaces/csp_if_sludp.h>

#ifdef HAVE_LIBZMQ
#include <zmq.h>
#include <csp/interfaces/csp_if_zmqhub.h>
#endif

#define SATCTL_PROMPT_FMT		"\033[94m[%s]\033[0m "
#define SATCTL_DEFAULT_ADDRESS		1
#define SATCTL_DEFAULT_TYPE		TYPE_NONE
#define SATCTL_DEFAULT_RATE		1000000
#define SATCTL_CSP_MTU			1024
#define SATCTL_CSP_BUFFERS		255

#define SATCTL_LINE_SIZE		128
#define SATCTL_HISTORY_SIZE		2048

#define SATCTL_DESC			"Satlab Control Application " SATCTL_VERSION "\n"
#define SATCTL_COPYRIGHT		"Copyright (c) 2014-2022 Satlab A/S <satlab@satlab.com>\n"

#define vmsg(_verbose, _fmt, ...) do { \
	if (_verbose) \
		printf("\033[1m" _fmt "\033[0m\n", ##__VA_ARGS__); \
	} while (0);

enum conntype {
	TYPE_NONE,
	TYPE_CAN,
	TYPE_UART,
	TYPE_MCAST,
	TYPE_UDP,
	TYPE_SLGND,
	TYPE_ZMQHUB,
};

static void usage(void)
{
	printf("usage: satctl [options] [command/file]\n");
	printf("\n");
	printf(SATCTL_DESC);
	printf(SATCTL_COPYRIGHT);
	printf("\n");
	printf("Options:\n");
	printf(" -c,\t\tUse CAN as transport interface\n");
	printf(" -m,\t\tUse MCAST as transport interface\n");
	printf(" -u,\t\tUse UART/KISS as transport interface\n");
	printf(" -s,\t\tUse SLGND as transport interface\n");
	printf(" -e,\t\tUse UDP as transport interface\n");
#ifdef HAVE_LIBZMQ
	printf(" -z,\t\tUse ZMQHUB as transport interface\n");
	printf(" -Z,\t\tStart built-in ZMQPROXY for ZMQHUB\n");
#endif
	printf(" -i ID,\t\tShow ID as prompt and CSP ident hostname\n");
	printf(" -d DEVICE,\tUse DEVICE as CAN/UART/MCAST/UDP interface or SLGND/ZMQ host[:port]\n");
	printf(" -r RATE,\tUse RATE bps as CAN/UART bitrate\n");
	printf(" -n NODE\tUse NODE as own CSP address\n");
	printf(" -g NODE\tUse NODE as gateway (via) CSP address\n");
	printf(" -k KEY\t\tUse KEY as CSP XTEA/HMAC key\n");
	printf(" -h,\t\tPrint this help and exit\n");
	printf(" -v,\t\tEnable verbose output\n");
}

#ifdef HAVE_LIBZMQ
struct zmqproxy_args {
	void *ctx;
	void *frontend;
	void *backend;
};

static void *thread_csp_zmqproxy(void *param)
{
	struct zmqproxy_args *args = param;

	zmq_proxy(args->frontend, args->backend, NULL);
	zmq_ctx_destroy(args->ctx);
	free(args);

	return NULL;
}

static int start_zmqproxy(void)
{
	pthread_t zmqproxy;
	struct zmqproxy_args *args;

	args = calloc(1, sizeof(*args));
	if (!args) {
		fprintf(stderr, "Failed to allocate zmqproxy args\n");
		return -ENOMEM;
	}

	args->ctx = zmq_ctx_new();
	if (!args->ctx) {
		fprintf(stderr, "Failed to create zmq context\n");
		goto err_free;
	}

	args->frontend = zmq_socket(args->ctx, ZMQ_XSUB);
	if (!args->frontend) {
		fprintf(stderr, "Failed to create zmqproxy frontend socket\n");
		goto err_free_ctx;
	}

	args->backend = zmq_socket(args->ctx, ZMQ_XPUB);
	if (!args->backend) {
		fprintf(stderr, "Failed to create zmqproxy backend socket\n");
		goto err_close_frontend;
	}

	if (zmq_bind(args->frontend, "tcp://*:6000") != 0) {
		fprintf(stderr, "Failed to bind zmqproxy frontend\n");
		goto err_close_backend;
	}

	if (zmq_bind(args->backend, "tcp://*:7000") != 0) {
		fprintf(stderr, "Failed to bind zmqproxy backend\n");
		goto err_close_backend;
	}

	if (pthread_create(&zmqproxy, NULL, thread_csp_zmqproxy, args) != 0) {
		fprintf(stderr, "Failed to start zmqproxy thread\n");
		goto err_close_backend;
	}

	return 0;

err_close_backend:
	zmq_close(args->backend);
err_close_frontend:
	zmq_close(args->frontend);
err_free_ctx:
	zmq_ctx_destroy(args->ctx);
err_free:
	free(args);

	return -EIO;
}
#endif

static void *thread_csp_service(void *param)
{
	csp_socket_t *sock;
	csp_conn_t *conn;
	csp_packet_t *packet;

	/* Unused */
	(void)param;

	sock = csp_socket(CSP_SO_NONE);
	csp_bind(sock, CSP_ANY);
	csp_listen(sock, 10);

	while (1) {
		if ((conn = csp_accept(sock, 10000)) == NULL)
			continue;

		packet = csp_read(conn, 1000);
		if (packet != NULL)
			csp_service_handler(conn, packet);

		csp_close(conn);
	}

	return NULL;
}

static int configure_csp(uint8_t addr, unsigned int type, const char *device, unsigned int rate, const char *key, const char *hostname, uint8_t via)
{
	int ret;
	csp_conf_t conf;
	csp_iface_t *ifc;

	csp_conf_get_defaults(&conf);
	conf.address = addr;
	conf.hostname = hostname;
	conf.model = "satctl";
	conf.revision = SATCTL_VERSION;
	conf.conn_queue_length = SATCTL_CSP_BUFFERS;
	conf.fifo_length = SATCTL_CSP_BUFFERS;
	conf.buffers = SATCTL_CSP_BUFFERS;
	conf.buffer_data_size = SATCTL_CSP_MTU;

	if (csp_init(&conf) < 0)
		return -1;

	if (type == TYPE_NONE) {
		ifc = &csp_if_lo;
	} else if (type == TYPE_CAN) {
		ret = csp_can_socketcan_open_and_add_interface(device ? device : "can0", CSP_IF_CAN_DEFAULT_NAME, 0, false, &ifc);
		if (ret != CSP_ERR_NONE)
			return -1;
	} else if (type == TYPE_UART) {
		csp_usart_conf_t usart_conf = {
			.device = device ? device : "/dev/ttyUSB0",
			.baudrate = rate,
			.databits = 8,
			.stopbits = 1,
			.paritysetting = 0,
			.checkparity = 0
		};
		ret = csp_usart_open_and_add_kiss_interface(&usart_conf, CSP_IF_KISS_DEFAULT_NAME, &ifc);
		if (ret != CSP_ERR_NONE)
			return -1;
	} else if (type == TYPE_MCAST) {
		if (!device)
			device = "localhost";
		if (csp_mcast_init(device, CSP_IF_MCAST_DEFAULT_NAME, &ifc) < 0)
			return -1;
	} else if (type == TYPE_UDP) {
		if (!device)
			device = "eth0";
		if (csp_sludp_init(device, CSP_IF_SLUDP_DEFAULT_NAME, &ifc) < 0)
			return -1;
	} else if (type == TYPE_SLGND) {
		if (!device)
			device = "localhost";
		if (csp_slgnd_init(device, CSP_IF_SLGND_DEFAULT_NAME, &ifc) < 0)
			return -1;
#ifdef HAVE_LIBZMQ
	} else if (type == TYPE_ZMQHUB) {
		if (!device)
			device = "localhost";
		if (csp_zmqhub_init(csp_get_address(), device, 0, &ifc) < 0)
			return -1;
#endif
	} else {
		fprintf(stderr, "Unknown interface\n");
		return -1;
	}

	if (csp_rtable_set(CSP_DEFAULT_ROUTE, 0, ifc, via) < 0)
		return -1;

	if (key) {
		if (csp_xtea_set_key(key, strlen(key)) < 0) {
			fprintf(stderr, "Failed to set XTEA key\n");
			return -1;
		}
		if (csp_hmac_set_key(key, strlen(key)) < 0) {
			fprintf(stderr, "Failed to set HMAC key\n");
			return -1;
		}
	}

	if (csp_route_start_task(0, 0) < 0)
		return -1;

	pthread_t csp_thread;
	pthread_create(&csp_thread, NULL, thread_csp_service, NULL);

	return 0;
}

int main(int argc, char **argv)
{
	struct slash *slash;
	bool verbose = false;
	int remain, index, i, c, p = 0;
	char *ex, *endptr;
	FILE * fp;
	ssize_t nread;
	size_t len = 0;
	char *line = NULL;
	char prompt[32];
	const char *name = NULL;
	char hostname[32] = "satctl";

	unsigned int type = SATCTL_DEFAULT_TYPE;
	uint8_t addr = SATCTL_DEFAULT_ADDRESS, via = CSP_NO_VIA_ADDRESS;
	char *device = NULL;
	unsigned int rate = SATCTL_DEFAULT_RATE;
	char *crypto_key = NULL;
#ifdef HAVE_LIBZMQ
	bool zmqproxy = false;
#endif

	while ((c = getopt(argc, argv, "+hcmueszZd:i:n:g:r:k:v")) != -1) {
		switch (c) {
		case 'h':
			usage();
			exit(EXIT_SUCCESS);
		case 'c':
			type = TYPE_CAN;
			break;
		case 'm':
			type = TYPE_MCAST;
			break;
		case 'u':
			type = TYPE_UART;
			break;
		case 'e':
			type = TYPE_UDP;
			break;
		case 's':
			type = TYPE_SLGND;
			break;
#ifdef HAVE_LIBZMQ
		case 'z':
			type = TYPE_ZMQHUB;
			break;
		case 'Z':
			zmqproxy = true;
			break;
#endif
		case 'd':
			device = optarg;
			break;
		case 'i':
			name = optarg;
			break;
		case 'n':
			addr = strtoul(optarg, &endptr, 0);
			if (*endptr != '\0' || addr < 1 || addr > CSP_ID_HOST_MAX) {
				fprintf(stderr, "Invalid CSP address\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 'g':
			via = strtoul(optarg, &endptr, 0);
			if (*endptr != '\0' || via < 1 || via > CSP_ID_HOST_MAX) {
				fprintf(stderr, "Invalid CSP gateway address\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 'r':
			rate = strtoul(optarg, &endptr, 0);
			if (*endptr != '\0') {
				fprintf(stderr, "Invalid rate\n");
				exit(EXIT_FAILURE);
			}
			break;
		case 'k':
			crypto_key = optarg;
			break;
		case 'v':
			verbose = true;
			break;
		case '?':
			if (optopt == 'c')
				fprintf(stderr, "Option -%c requires an argument\n", optopt);
			/* FALLTHROUGH */
		default:
			exit(EXIT_FAILURE);
		}
	}

	remain = argc - optind;
	index = optind;

	/* Build prompt and hostname */
	if (name) {
		strncpy(hostname, name, sizeof(hostname) - 1);
		hostname[sizeof(hostname) - 1] = '\0';
	} else {
		name = "satctl";
		gethostname(hostname, sizeof(hostname));
		hostname[sizeof(hostname) - 1] = '\0';
	}

	snprintf(prompt, sizeof(prompt) - 1, SATCTL_PROMPT_FMT, name);
	prompt[sizeof(prompt) - 1] = '\0';

#ifdef HAVE_LIBZMQ
	if (zmqproxy && start_zmqproxy() < 0) {
		fprintf(stderr, "Failed to start ZMQPROXY\n");
		exit(EXIT_FAILURE);
	}
#endif

	if (configure_csp(addr, type, device, rate, crypto_key, hostname, via) < 0) {
		fprintf(stderr, "Failed to init CSP\n");
		exit(EXIT_FAILURE);
	}

	slash = slash_create(SATCTL_LINE_SIZE, SATCTL_HISTORY_SIZE);
	if (!slash) {
		fprintf(stderr, "Failed to init slash\n");
		exit(EXIT_FAILURE);
	}

	/* Interactive or one-shot mode */
	if (remain > 0) {
		/* Try to open script file */	
		fp = fopen(argv[index], "r");
		if (fp != NULL) {
			if (remain > 1) {
				fprintf(stderr, "Only one file argument is allowed\n");
				exit(EXIT_FAILURE);
			}

			while ((nread = getline(&line, &len, fp)) != -1) {
				if (line[nread - 1] == '\n') {
					line[nread - 1] = '\0';
					nread--;
				}

				vmsg(verbose, "Running command \'%s\':", line);

				if (slash_execute(slash, line) < 0)
					exit(EXIT_FAILURE);
			}

			fclose(fp);
			free(line);
		} else {
			/* Build command string */
			ex = calloc(SATCTL_LINE_SIZE, 1);
			if (!ex) {
				fprintf(stderr, "Failed to allocate command memory");
				exit(EXIT_FAILURE);
			}

			for (i = 0; i < remain; i++) {
				if (i > 0)
					p = ex - strncat(ex, " ", SATCTL_LINE_SIZE - p);
				p = ex - strncat(ex, argv[index + i], SATCTL_LINE_SIZE - p);
			}

			vmsg(verbose, "Running command \'%s\':", ex);

			if (slash_execute(slash, ex) < 0)
				exit(EXIT_FAILURE);

			free(ex);
		}
	} else {
		printf(SATCTL_DESC);
		printf(SATCTL_COPYRIGHT "\n");

		if (type == TYPE_NONE)
			printf("No interface selected, loopback only mode!\n\n");

		printf("Run 'help' for a list of available commands\n\n");

		slash_loop(slash, prompt, NULL);
	}

	slash_destroy(slash);

	return 0;
}
